/*
 * Author: Michael Jonas (copyright 2020) released under GNU GPL v3
 * Date: 3/18/20
 * ZXSimulator: a prototype ZX80/81 BASIC Simulator
 */


#include stdio_h


/****************************** cmd_h *******************************/

#define QLBREAK   ((keyrow(7)==2) && (keyrow(2)==8))
#define QLSTOP    ((keyrow(7)==2) && (keyrow(3)==8))
#define QLGO      ((keyrow(7)==2) && (keyrow(6)==8))

#define addr       long

#define EOL       '\n'
#define EOP       '\0'
#define LINELEN    256
#define CMDLEN     20
#define STRLEN     80
#define VARLEN     20
#define BLKLEN     64
#define COLLEN     72
#define NUMLEN     10
#define MAXVAR     7     /* needs to be prime (i.e. 257) */
#define MAXLNUM    9999

#define MAXNEST    10
#define CMDLINE    1
#define NOLP       2
#define FORLP      3
#define REPEATLP   4

#define ENDIF      0
#define ELSEIF     1

#define NOERR      0
#define NOCMD     -1
#define NOVAL     -2
#define NOHEAP    -3
#define MAXLIM    -4
#define BADARG    -5
#define BADCHAR   -6
#define NOQUOTE   -7
#define UNKNOWN   -8

#define MAXCOL     440
#define MAXROW     124
#define LASTCOL    428
#define LASTROW    108
#define SIZE       5


/* global command line variables */
int  globerr = NOERR;
char pre_ws[6] = {':',' ','(',')','\n','\0'};
char pos_ws[6] = {'-',' ','(',')','\n','\0'};
char arg_dl[3] = {'\"','\'','\0'};
char cstr[LINELEN] = "";
char *pstr;
int  plen = 0;
char *str;
int  sp = 0;


char mask[8] = { 128, 64, 32, 16, 8, 4, 2, 1 };

char chQUO[SIZE] = { 0x24, 0x90, 0x00, 0x00, 0x00 };
char chPOU[SIZE] = { 0x1C, 0x8F, 0x88, 0x21, 0xF8 };
char chDOL[SIZE] = { 0x08, 0xFA, 0x8F, 0x8A, 0xF9 };
char chCOL[SIZE] = { 0x00, 0x01, 0x00, 0x00, 0x40 };
char chQUE[SIZE] = { 0x3D, 0x08, 0x42, 0x00, 0x20 };
char chOPA[SIZE] = { 0x04, 0x20, 0x82, 0x08, 0x10 };
char chCPA[SIZE] = { 0x20, 0x41, 0x04, 0x10, 0x80 };
char chGRE[SIZE] = { 0x00, 0x40, 0x81, 0x08, 0x40 };
char chLES[SIZE] = { 0x00, 0x10, 0x84, 0x08, 0x10 };
char chEQU[SIZE] = { 0x00, 0x03, 0xE0, 0x3E, 0x00 };
char chPLU[SIZE] = { 0x00, 0x20, 0x8F, 0x88, 0x20 };
char chMIN[SIZE] = { 0x00, 0x00, 0x0F, 0x80, 0x00 };
char chUND[SIZE] = { 0x00, 0x00, 0x00, 0x00, 0xF8 };
char chAST[SIZE] = { 0x00, 0x50, 0x8F, 0x88, 0x50 };
char chDIV[SIZE] = { 0x00, 0x08, 0x42, 0x10, 0x80 };
char chSEM[SIZE] = { 0x00, 0x40, 0x00, 0x10, 0x44 };
char chCOM[SIZE] = { 0x00, 0x00, 0x00, 0x08, 0x22 };
char chPER[SIZE] = { 0x00, 0x00, 0x00, 0x18, 0x60 };
char chSPA[SIZE] = { 0x00, 0x00, 0x00, 0x00, 0x00 };

char ch0[SIZE] = { 0x3D, 0x1C, 0xB4, 0xE2, 0xF0 };
char ch1[SIZE] = { 0x18, 0xA0, 0x82, 0x08, 0xF8 };
char ch2[SIZE] = { 0x3D, 0x08, 0x2F, 0x41, 0xF8 };
char ch3[SIZE] = { 0x3D, 0x08, 0xC0, 0xC2, 0xF0 };
char ch4[SIZE] = { 0x08, 0x62, 0x92, 0x7E, 0x20 };
char ch5[SIZE] = { 0x7F, 0x07, 0xC0, 0xC2, 0xF0 };
char ch6[SIZE] = { 0x3D, 0x07, 0xD0, 0xC2, 0xF0 };
char ch7[SIZE] = { 0x7E, 0x08, 0x42, 0x10, 0x40 };
char ch8[SIZE] = { 0x3D, 0x0B, 0xD0, 0xC2, 0xF0 };
char ch9[SIZE] = { 0x3D, 0x0C, 0x2F, 0x82, 0xF0 };

char chA[SIZE] = { 0x3D, 0x0C, 0x3F, 0xC3, 0x08 };
char chB[SIZE] = { 0x7D, 0x0F, 0xD0, 0xC3, 0xF0 };
char chC[SIZE] = { 0x3D, 0x0C, 0x10, 0x42, 0xF0 };
char chD[SIZE] = { 0x79, 0x14, 0x30, 0xC5, 0xE0 };
char chE[SIZE] = { 0x7F, 0x07, 0xD0, 0x41, 0xF8 };
char chF[SIZE] = { 0x7F, 0x07, 0xD0, 0x41, 0x00 };
char chG[SIZE] = { 0x3D, 0x0C, 0x13, 0xC2, 0xF0 };
char chH[SIZE] = { 0x43, 0x0F, 0xF0, 0xC3, 0x08 };
char chI[SIZE] = { 0x3E, 0x20, 0x82, 0x08, 0xF8 };
char chJ[SIZE] = { 0x02, 0x08, 0x30, 0xC2, 0xF0 };
char chK[SIZE] = { 0x45, 0x27, 0x12, 0x45, 0x08 };
char chL[SIZE] = { 0x41, 0x04, 0x10, 0x41, 0xF8 };
char chM[SIZE] = { 0x43, 0x9D, 0xB0, 0xC3, 0x08 };
char chN[SIZE] = { 0x43, 0x8D, 0x32, 0xC7, 0x08 };
char chO[SIZE] = { 0x3D, 0x0C, 0x30, 0xC2, 0xF0 };
char chP[SIZE] = { 0x7D, 0x0C, 0x3F, 0x41, 0x00 };
char chQ[SIZE] = { 0x3D, 0x0C, 0x34, 0xCA, 0xF0 };
char chR[SIZE] = { 0x7D, 0x0C, 0x3F, 0x45, 0x08 };
char chS[SIZE] = { 0x3D, 0x03, 0xC0, 0xC2, 0xF0 };
char chT[SIZE] = { 0xFE, 0x41, 0x04, 0x10, 0x40 };
char chU[SIZE] = { 0x43, 0x0C, 0x30, 0xC2, 0xF0 };
char chV[SIZE] = { 0x43, 0x0C, 0x30, 0xA4, 0x60 };
char chW[SIZE] = { 0x43, 0x0C, 0x30, 0xDB, 0x98 };
char chX[SIZE] = { 0x42, 0x91, 0x86, 0x25, 0x08 };
char chY[SIZE] = { 0x83, 0x12, 0x84, 0x10, 0x40 };
char chZ[SIZE] = { 0x7E, 0x10, 0x84, 0x21, 0xF8 };


/* data sturctures for variables */
addr var_nam[MAXVAR] = {0L,0L,0L,0L,0L,0L,0L};
addr var_str[MAXVAR] = {0L,0L,0L,0L,0L,0L,0L};
int  var_num[MAXVAR] = {0,0,0,0,0,0,0};

/* data structures for loops gosubs, and if statements */
int  lp_ptr[MAXNEST] = {0,0,0,0,0,0,0,0,0,0};
int  lp_typ[MAXNEST] = {0,0,0,0,0,0,0,0,0,0};
int  lp_end[MAXNEST] = {0,0,0,0,0,0,0,0,0,0};
int  lp_inc[MAXNEST] = {0,0,0,0,0,0,0,0,0,0};
addr lp_var[MAXNEST] = {0L,0L,0L,0L,0L,0L,0L,0L,0L,0L};
int  lp_ind = -1;

int  rt_ptr[MAXNEST] = {0,0,0,0,0,0,0,0,0,0};
addr rt_str[MAXNEST] = {0L,0L,0L,0L,0L,0L,0L,0L,0L,0L};
int  rt_ind = -1;

int  if_nst = 0;

int  col = 0;
int  row = 0;


/****************************** cmd_c *******************************/

main ()
{
   int ln;
   int sRow, sCol;

   initPARSE();

   while (1)
   {
      reset(); globerr = NOERR;

      ZXgets(stdout, cstr, LINELEN);

      if (linenum(&ln) == NOERR)
         { if (addline(ln) != NOERR) showerr("BAD LINE NUMBER"); }
      else switch (PARSEcmds())
      {
         case NOERR  : break;
         case NOCMD  : showerr("NO COMMAND"); break;
         case BADARG : showerr("BAD ARGUMENT"); break;
         default     : showerr("UNKNOWN");
      }
   }
}   /* end main */


/* initializes the parser */
initPARSE ()
{
   initQLscr();
   str = cstr;
   plen = 0;
   setpstr(0,EOL); setpstr(1,EOP);
}   /* end initPARSE */


/* parser */
PARSEcmds()
{
   char cmd[CMDLEN];
   int  err, ln;

   while (1)
   {
      if (str[sp] == ':')
         sp++;
      else if (str[sp] == EOL)
      {
         if (lp_ind >= 0) if (lp_typ[lp_ind] == CMDLINE)
            if (nextlp(lp_var[lp_ind]) != NOERR) return UNKNOWN;
         if (str[sp++] == EOL) linenum(&ln);
      }

      if (str[sp] == EOP) return NOERR;
      if (QLBREAK) return NOERR;
      if (QLSTOP) while (!QLGO);

      if (getcmd(cmd) != NOERR) return BADARG;

      if ((err=BASICcmds(cmd)) != NOCMD)
      {
         if (err != NOERR)
            return err;
      }
   }
}   /* end PARSEcmds */


/* BASIC commands */
BASICcmds (cmd)
   char cmd[];
{
   char *val_s, *vl2_s, arg[LINELEN], var[VARLEN], *val_s, *vl2_s;
   int  val_i, vl2_i, vl3_i, i, typ;

   if (streql("BEEP",cmd))
   {
      if (noarg() == NOERR)
         beep(1,1);
      else
      {
         if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
         if (noarg() != NOERR) return BADARG;
         for (i=0; i<val_i; i++) beep(1,1);
      }
   }
   else if (streql("CLS",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      cls(stdout);
      row = col = 0;
   }
   else if (streql("EDIT",cmd))
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (noarg() != NOERR) return BADARG;
      if (editline(val_i) != NOERR) return BADARG;
   }
   else if (streql("ELSE",cmd))
   {
      if (noarg() != NOERR)
         skipln();
      else if (!if_nst)
         return BADARG;
      else if (endif(ENDIF) == NOERR)
         if_nst--;
      else
         return BADARG;
   }
   else if (streql("END",cmd))
   {
      if (getkey("FOR") == NOERR)
         { if (getvar(var) != NOERR) return BADARG;
           if (noarg() != NOERR) return BADARG;
           if (nextlp(var) != NOERR) return BADARG; }
      else if (getkey("IF") == NOERR)
         { if (noarg() != NOERR) return BADARG;
           if (if_nst) if_nst--; else return BADARG; }
      else if (getkey("REPEAT") == NOERR)
         { if (getvar(var) != NOERR) return BADARG;
           if (noarg() != NOERR) return BADARG;
           if (nextlp(var) != NOERR) return BADARG; }
      else
         return BADARG;
   }
   else if (streql("EXIT",cmd))
   {
      if (getvar(var) != NOERR) return BADARG;
      if (noarg() != NOERR) return BADARG;
      if (exitlp(var) != NOERR) return BADARG;
   }
   else if (streql("FOR",cmd))
   {
      if (getvar(var) != NOERR) return BADARG;
      if (getsym("=") != NOERR) return BADARG;
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (getkey("TO") != NOERR) return BADARG;
      if (getexpr(&vl2_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (getkey("STEP") != NOERR)
         vl3_i = 1;
      else if (getexpr(&vl3_i,&val_s) == NOERR)
         free(val_s);
      else
         return BADARG;
      if (noarg() != NOERR) return BADARG;
      if (str[sp] == ':') typ = CMDLINE; else typ = FORLP;
      if (nestlp(var,val_i,vl2_i,vl3_i,typ) != NOERR) return BADARG;
   }
   else if (streql("GOSUB",cmd))
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (noarg() != NOERR) return BADARG;
      if ((++rt_ind) >= MAXNEST) return BADARG;
      rt_ptr[rt_ind] = sp; rt_str[rt_ind] = str;
      findline(val_i);
   }
   else if (streql("GOTO",cmd))
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (noarg() != NOERR) return BADARG;
      findline(val_i);
   }
   else if (streql("IF",cmd))
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (getkey("THEN") != NOERR) return BADARG;
      if (noarg() != NOERR)
         { if (!val_i) findkey("ELSE",EOL); }
      else if (val_i)
         if_nst++;
      else if (endif(ELSEIF) != NOERR)
         return BADARG;
      else if (toupper(str[sp-1]) == 'E')  /* ELS'E' */
         if_nst++;
   }
   else if (streql("INPUT",cmd))
   {
      if (getvar(var) != NOERR) return BADARG;
      if (noarg() != NOERR) return BADARG;

      fgets(arg,LINELEN,stdin); arg[strlen(arg)-1]='\0';
      storevar(var,atoi(arg),arg);
   }
   else if (streql("LET",cmd))
   {
      if (getvar(var) != NOERR) return BADARG;
      if (getsym("=") != NOERR)
         { fetchvar(var,&val_i,&val_s);
           if (getrange(&vl2_i,&vl3_i,strlen(val_s)) != NOERR) { free(val_s); return BADARG; }
           if (getsym("=") != NOERR) { free(val_s); return BADARG; }
           if (getexpr(&val_i,&vl2_s) != NOERR) { free(val_s); return BADARG; }
           for (i=vl2_i-1; i<vl3_i; i++)
              if (vl2_s[i-vl2_i+1]=='\0') break; else val_s[i]=vl2_s[i-vl2_i+1];
           val_i=atoi(val_s); free(vl2_s); }
      else
         { if (getexpr(&val_i,&val_s) != NOERR) return BADARG; }

      if (noarg() != NOERR) { free(val_s); return BADARG; }
      storevar(var,val_i,val_s); free(val_s);
   }
   else if (streql("LIST",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      writestr(pstr,stdout);
   }
   else if (streql("LOAD",cmd))
   {
      if (getfname(arg) != NOERR) return BADARG;
      if (noarg() != NOERR) return BADARG;
      if (loadfile(arg) != NOERR) return BADARG;
   }
   else if (streql("LRUN",cmd))
   {
      if (getfname(arg) != NOERR) return BADARG;
      if (noarg() != NOERR)
         { if (getsym(",") != NOERR) return BADARG;
           if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
           if (noarg() != NOERR) return BADARG;
           if (loadfile(arg) != NOERR) return BADARG;
           reset(); findline(val_i); }
      else
         { if (loadfile(arg) != NOERR) return BADARG;
           reset(); str = pstr; }
   }
   else if (streql("NEW",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      reset(); setpstr(0,EOL); setpstr(1,EOP);
   }
   else if (streql("NEXT",cmd))
   {
      if (getvar(var) != NOERR) return BADARG;
      if (noarg() != NOERR) return BADARG;
      if (nextlp(var) != NOERR) return BADARG;
   }
   else if (streql("PRINT",cmd))
   {
      if (putformat() != NOERR) return BADARG;
   }
   else if (streql("REM",cmd))
   {
      skipln();
   }
   else if (streql("REPEAT",cmd))
   {
      if (getvar(var) != NOERR) return BADARG;
      if (noarg() != NOERR) return BADARG;
      if (str[sp] == ':') typ = CMDLINE; else typ = REPEATLP;
      if (nestlp(var,0,0,0,typ) != NOERR) return BADARG;
   }
   else if (streql("RETURN",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      if (rt_ind < 0) return BADARG;
      sp = rt_ptr[rt_ind]; str = rt_str[rt_ind--];
   }
   else if (streql("RUN",cmd))
   {
      if (noarg() != NOERR)
         { if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
           if (noarg() != NOERR) return BADARG;
           reset(); findline(val_i); }
      else
         { reset(); str = pstr; }
   }
   else if (streql("SAVE",cmd))
   {
      if (getfname(arg) != NOERR) return BADARG;
      if (noarg() != NOERR) return BADARG;
      if (savefile(arg) != NOERR) return BADARG;
   }
   else if (streql("STOP",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      reset();
   }
   else if (streql("QUIT",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      abort(0);
   }
   else if (cmd[0] != '\0')        /* if command is not empty */
      return NOCMD;

   return NOERR;
}   /* end BASICcmds */


/****************************** aux_c *******************************/

/* gets command from command line */
getcmd (cmd)
   char cmd[];
{
   char ch;
   int  i;

   while (str[sp] == ' ') sp++;
   for (i=0; i<CMDLEN; i++,sp++)
      if (((ch=str[sp]) != ' ') && (ch != ':') && (ch != '\n'))
         cmd[i] = toupper(ch);
      else
        { cmd[i] = '\0'; return NOERR; }
   return BADARG;
}   /* end getcmd */


/* gets symbol from command line */
getsym (sym)
   char sym[];
{
   int i, op;

   op = sp;
   while (str[sp] == ' ') sp++;
   for (i=0; sym[i]!='\0'; i++)
      if (toupper(str[sp]) == sym[i]) sp++; else { sp=op; return BADCHAR; }
   return NOERR;
}   /* getsym */


/* gets keyword from command line */
getkey (key)
   char key[];
{
   char ch;
   int i, tp;

   tp = sp;
   while (str[tp] == ' ') tp++;
   if (strchr(pre_ws,str[tp-1]) == NULL) return BADCHAR;
   for (i=0; key[i]!='\0'; i++)
      if (toupper(str[tp]) == key[i]) tp++; else return BADCHAR;
   if (strchr(pos_ws,str[tp]) == NULL) return BADCHAR;
   sp = tp;
   return NOERR;
}   /* getkey */


/* finds keyword from command line */
findkey (key, ech)
   char key[], ech;
{
   char ch;

   ch = key[0];

   while (str[sp] != ech)
   {
      if (toupper(str[sp]) == ch) if (getkey(key) == NOERR) return NOERR;
      sp++;
   }

   return BADARG;
}   /* findkey */


/* finds keyword from command line within range bp & ep */
instrkey (key, bp, ep)
   char key[];
   int  bp, ep;
{
   char ch;

   ch = key[0];

   for (sp=bp; sp<ep; sp++)
      if (toupper(str[sp]) == ch) if (getkey(key) == NOERR) return NOERR;

   return BADARG;
}   /* instrkey */


/* displays formatted argument for print statement */
putformat ()
{
   int  val_i, doLF;
   char *val_s;

   doLF = 1;
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case ','  : { sp++; doLF = 0; ZXprintf(stdout,"         "); break; }
         case ';'  : { sp++; doLF = 0; break; }
         case ':'  :
         case '\n' : { if (doLF) ZXprintf(stdout,"\n"); return NOERR; }
         default   : {
                       if (getexpr(&val_i,&val_s) != NOERR) return BADARG;
                       doLF = 1; ZXprintf(stdout, val_s); free(val_s);
                     }
      }
   }
}   /* end putformat */


/* gets next argument from command line */
nextarg (arg)
   char arg[];
{
   int i;
   char dl, ch;

   while (str[sp] == ' ') sp++;
   if (str[sp-1] != ' ') return BADARG;
   if (strchr(arg_dl,str[sp]) == NULL) dl = ' '; else dl = str[sp++];

   for (i=0; i<LINELEN; i++,sp++)
      if (((ch=str[sp])!=dl) && (ch!='\n') && ((dl!=' ') || (ch!=':')))
         arg[i] = ch;
      else
         { arg[i] = '\0';
           if (dl != ' ') if (dl == ch) sp++; else return NOQUOTE;
           if (i) return NOERR; else return NOVAL; }

   return BADARG;
}   /* end nextarg */


/* gets last argument from command line */
lastarg (arg)
   char arg[];
{
   int err;

   if ((err=nextarg(arg)) != NOERR) return err;
   else if ((err=noarg()) != NOERR) return err;
   else return NOERR;
}   /* end lastarg */


/* checks for no argument from command line */
noarg ()
{
   while (str[sp] == ' ') sp++;
   if ((str[sp] != ':') && (str[sp] != '\n'))
      return BADCHAR;
   else
      return NOERR;
}   /* end noarg */


/* skips to end of line */
skipln ()
{
   while (str[sp] != '\n') sp++;
}   /* end skipln */


/* jumps to END IF (or ELSE if it exists) */
endif (doELSE)
   int doELSE;
{
   int ip, op, tp;

   op = sp; tp = sp;
   while (findkey("END",EOP) == NOERR)
      if (getkey("IF") == NOERR)
      {
         ip = sp; sp = tp;
         while (sp < ip)
            if (instrkey("THEN",tp,ip) != NOERR)
            {
               if (doELSE) instrkey("ELSE",tp,ip);
               return NOERR;
            }
            else
            {
               if (doELSE) if (instrkey("ELSE",tp,sp) == NOERR) return NOERR;
               if (noarg() != NOERR)
                  skipln();
               else if (endif(ENDIF) != NOERR)
                  { sp = op; return BADARG; }
               tp = sp;
            }
      }
   sp = op; return BADARG;
}   /* end endif */


/* jumps to end of loop */
endlp (lp)
   int lp;
{
   int os;

   os = sp;
   switch (lp_typ[lp])
   {
      case CMDLINE : {
         skipln();
         return NOERR;
      }
      case FORLP : {
         while (findkey("END",EOP) == NOERR)
            if (getkey("FOR") == NOERR) if (getkey(lp_var[lp]) == NOERR) return NOERR;
         break;
      }
      case REPEATLP : {
         while (findkey("END",EOP) == NOERR)
            if (getkey("REPEAT") == NOERR) if (getkey(lp_var[lp]) == NOERR) return NOERR;
         break;
      }
   }
   sp = os;
   return BADARG;
}   /* end endlp */


/* clears upto "st" looping variables */
clearlp (st)
   int st;
{
   int i;

   for (i=st; i<=lp_ind; i++)
      { lp_ptr[i]=0; lp_typ[i]=0; lp_end[i]=0; lp_inc[i]=0; free(lp_var[i]); lp_var[i]=0L; }
   lp_ind = st-1;
}   /* end clearlp */


/* exits all loops up to "var" */
exitlp (var)
   char var[];
{
   int  i;

   if (var[strlen(var)-1] == '$') return BADARG;

   for (i=lp_ind; i>=0; i--)
      if (streql(lp_var[i],var))
      {
         if (endlp(i) == NOERR) clearlp(i); else return BADARG;
         return NOERR;
      }
   return BADARG;
}   /* end exitlp */


/* skips to next increment of "var" */
nextlp (var)
   char var[];
{
   addr val_s;
   int i, val_i, inc, end;

   if (var[strlen(var)-1] == '$') return BADARG;

   for (i=lp_ind; i>=0; i--)
      if (streql(lp_var[i],var))
      {
         fetchvar(var,&val_i,&val_s); free(val_s);
         inc = lp_inc[i];  end = lp_end[i];
         val_i += inc;
         if (((inc > 0) && (val_i > end)) || ((inc < 0) && (val_i < end)))
            { clearlp(i);
              if (i) if (lp_typ[i-1] == CMDLINE)
                 if (nextlp(lp_var[i-1]) != NOERR) showerr("UNKNOWN"); }
         else
            { storevar(var,val_i,""); clearlp(i+1); sp=lp_ptr[i]; }
         return NOERR;
      }
   return BADARG;
}   /* end nextlp */


/* adds new loop nest for "var" */
nestlp (var,bgn,end,inc,typ)
   char var[];
   int bgn, end, inc, typ;
{
   addr mem;
   int  l;

   if ((lp_ind+1) == MAXNEST) return BADARG;
   if (var[strlen(var)-1] == '$') return BADARG;

   storevar(var,bgn,"");

   if (((inc > 0) && (bgn > end)) || ((inc < 0) && (bgn < end)))
      return BADARG;

   if (lp_var[++lp_ind] != 0L) free(lp_var[lp_ind]);
   if ((mem=malloc(strlen(var)+1)) == NULL)
      syserr(NOHEAP);
   else
   {
      lp_ptr[lp_ind] = sp;
      lp_typ[lp_ind] = typ;
      lp_end[lp_ind] = end;
      lp_inc[lp_ind] = inc;
      lp_var[lp_ind] = mem; strcpy(lp_var[lp_ind],var);
   }
      
   return NOERR;
}   /* end nestlp */


/* write string to specified device */
writestr (fstr,fp)
   char fstr[]; int fp;
{
   int i, j;
   str[LINELEN];

   j = 0;
   for (i=1; fstr[i]!=EOP; i++)
   {
      if (QLBREAK) return;
      if (QLSTOP) while (!QLGO);

      str[j++] = fstr[i];

         
      if (fstr[i] == '\n')
      {
         str[j] = '\0';
         ZXprintf(fp, str);
         j = 0;
      }
   }
}   /* end writestr */


/* saves program to file */
savefile (fname)
   char fname[];
{
   int fp, i;

   if ((fp=fopen(fname,"w")) == NULL) return BADARG;
   writestr(pstr,fp); fclose(fp);
   return NOERR;
}   /* end savefile */


/* loads file into string */
loadfile (fname)
   char fname[];
{
   int fp, l, ol, ln, ch, err;

   if ((fp=fopen(fname,"r")) == NULL) return BADARG;
   l = 0; ln = 0; setpstr(0,'\n');
   while ((err=checkline(fp,&ln,&l)) == NOERR)
      if ((err=readline(fp,&l)) != NOERR) break;

   if (err != BADCHAR)
      { if (pstr[l] != '\n') setpstr(++l,'\n'); setpstr(++l,'\0'); }
   else
      { ol=l; readline(fp,&l); pgmerr("BAD LINE",ol); setpstr(1,'\0'); }
   fclose(fp);
   return NOERR;
}   /* end loadfile */


/* reads character into program string, realloc's memory */
setpstr (l,ch)
   int l;
   char ch;
{
   char *ostr, *pmem;
   int  i;

   if (l >= plen)
   {
      if ((pmem=malloc(l+BLKLEN)) == NULL) { syserr(NOHEAP); return; }
      ostr = pstr; pstr = pmem;
      for (i=0; i<l; i++) pstr[i] = ostr[i];
      if (plen) free(ostr);
      plen = l + BLKLEN;
   }
   pstr[l] = ch;
}    /* end setpstr */


/* checks line number from file */
checkline (fp,ln,l)
   int fp, *ln, *l;
{
   char lnum[NUMLEN];
   int  i, num, ch, err;

   i = 0;
   while (((ch=fgetc(fp)) == ' ') || (ch == '\n'));
   if (ch == EOF) return EOF;
   while (isdigit(ch))
   {
      lnum[i++] = ch; setpstr(++(*l),ch);
      if ((ch=fgetc(fp)) == EOF) { (*l) -= i; return EOF; }
   }

   lnum[i] = '\0'; num = atoi(lnum); setpstr(++(*l),ch);
   if ((ch == ' ') && (num > *ln) && (num <= MAXLNUM))
      { *ln = num; return NOERR; }
   else
      return BADCHAR;
}   /* checkline */


/* reads line from file */
readline (fp,l)
   int fp, *l;
{
   int ch;

   if (pstr[*l] != '\n')
   {
      while ((ch=fgetc(fp)) != '\n')
         if (ch == EOF) return EOF; else setpstr(++(*l),ch);
      setpstr(++(*l),'\n');
   }
   return NOERR;
}   /* readline */


/* finds line number or gives error */
findline (num)
   int num;
{
   int lnum;

   sp = 0; str = pstr;
   while (str[sp] != EOP)
      if (str[sp++] == EOL)
         if (linenum(&lnum) == NOERR)
         {
            if (num < lnum)
               return NOVAL;
            else if (num == lnum)
               return NOERR;
         }
   return NOVAL;
}   /* findline */


/* gets line from program */
getline (lp, bp, ep)
   int lp, *bp, *ep;
{
   char ch;

   *bp=lp; while (*bp) if (pstr[(*bp)-1] != EOL) (*bp)--; else break;
   *ep=lp; while ((ch=pstr[*ep]) != EOP) if (ch != EOL) (*ep)++; else break;
}   /* getline */


/* gets line number from command line */
linenum (ln)
   int * ln;
{
   char num[NUMLEN];
   int  i;

   i=0;
   while(str[sp] == ' ') sp++;
   while(isdigit(str[sp])) num[i++]=str[sp++];
   num[i] = '\0'; *ln = atoi(num);
   if (i) return NOERR; else return NOVAL;
}   /* end linenum */


/* adds new line to program */
addline (num)
   int num;
{
   int err, cl, bp, cp, ep, np;

   if ((num < 1) || (num > MAXLNUM)) return BADARG;
   cp = sp; while (cstr[cp] == ' ') cp++;

   /* first delete line if found */
   err = findline(num); getline(sp,&bp,&ep);
   if (err == NOERR)
      { np=bp; while (pstr[ep] != EOP) pstr[np++] = pstr[++ep]; }
   else
      ep = strlen(pstr);

   /* if new line is not empty, add it */
   if (cstr[cp] != EOL)
   {
      np=0; while ((cstr[np]==' ') || (cstr[np]=='0')) np++;
      cl=strlen(cstr)-np; while (ep >= bp) setpstr(ep+cl, pstr[ep--]);
      for (cp=np; cstr[cp]!='\0'; cp++) setpstr(bp++, cstr[cp]);
   }

   reset(); return NOERR;
}   /* end addline */


/* edits line of program */
editline (num)
   int num;
{
   char mstr[LINELEN], *ostr;
   int  i, bp, ep, op;

   op = sp; ostr = str;
   if (findline(num) == NOERR)
   {
      getline(sp,&bp,&ep);
      for (i=bp; i<ep; i++) mstr[i-bp] = str[i];
      mstr[ep-bp] = '\0';
      ZXprintf(stdout, "EDIT: \n");
      reset(); return NOERR;
   }
   sp = op; str = ostr;
   return NOVAL;
}   /* end editline */


/* copy var to file */
tovar (fname,var)
   char fname[], var[];
{
   char val_s[];
   int  i, j, fp;

   if ((fp=fopen(fname,"r")) == NULL) return BADARG;
   j=0; while (fgetc(fp) != EOF) j++;
   memarg(&val_s,j+1); rewind(fp);
   for (i=0; i<j; i++) val_s[i] = fgetc(fp);
   val_s[i] = '\0'; storevar(var,atoi(val_s),val_s);
   free(val_s); fclose(fp);
   return NOERR;
}   /* end tovar */


/* copy var to file */
fromvar (var,fname)
   char var[], fname[];
{
   char val_s[], ch;
   int  val_i, i, fp;

   if ((fp=fopen(fname,"w")) == NULL) return BADARG;
   fetchvar(var,&val_i,&val_s);
   for (i=0; (ch=val_s[i])!='\0'; i++) ZXputc(fp, ch, 0, 7);
   free(val_s); fclose(fp);
   return NOERR;
}   /* end fromvar */


/* get file name */
getfname (fname)
   char fname[];
{
   char val_s[];
   int  val_i, ap;

   while (str[sp] == ' ') sp++;
   if (str[sp-1] == ' ') ap = sp; else return BADARG;

   if (evalcat(&val_i,&val_s) != NOERR)
      { sp = ap; return nextarg(fname); }
   else if (val_s[0] == '\0')
      { free(val_s); sp = ap; return nextarg(fname); }
   else
      { strcpy(fname,val_s); free(val_s); return NOERR; }
}   /* end getfname */


/* points to error in program line */
pgmerr (msg,lp)
   char msg[];
   int  lp;
{
   char estr[LINELEN];
   int  i, bp, ep;

   getline(lp,&bp,&ep); for (i=bp; i<ep; i++) estr[i-bp] = pstr[i];
   estr[ep-bp] = '\n'; estr[ep-bp+1] = '\0';
   printerr(msg,estr,lp-bp);
}   /* end pgmerr */


/* points to error in command line */
showerr (msg)
   char msg[];
{
   if (str == cstr)
      printerr(msg,"",sp);
   else
      pgmerr(msg,sp);
   reset();
}   /* end showerr */


/* displays error message */
printerr (msg,estr,ep)
   char msg[], estr[];
   int  ep;
{
   char mstr[STRLEN];
   int i, pt;

   ep = ep % COLLEN;
   if (ep > (COLLEN/2)) pt=ep-strlen(msg)-3; else pt=ep;
   for(i=0; i<pt; i++)
      mstr[i] = ' ';
   mstr[pt] = '\0';

   if (pt == ep)
   {
      ZXprintf(stdout, estr);
      ZXprintf(stdout, mstr);
      ZXprintf(stdout, "--- ");
      ZXprintf(stdout, msg);
      ZXprintf(stdout, "\n");
   }
   else
   {
      ZXprintf(stdout, estr);
      ZXprintf(stdout, mstr);
      ZXprintf(stdout, msg);
      ZXprintf(stdout, " ---\n");
   }
}   /* end printerr */


/* dynamic memory allocation of arg */
memarg(arg,len)
   addr *arg; int len;
{
   addr mem;

   if (globerr == NOERR)
      if (len > 0)
      {
         if ((mem=malloc(len)) == NULL)
            syserr(NOHEAP);
         else
            *arg = mem;
      }
}   /* end memarg */


/* sets char ptr to ch */
setchp(chp,ind,ch)
   char chp[]; int ind; char ch;
{
   chp[ind] = ch;
}   /* end setchp */


/* gets ch from char ptr */
getchp(chp,ind)
   char chp[]; int ind;
{
   return chp[ind];
}   /* end getchp */


/* gives system error */
syserr(err)
   int err;
{
   int ew;

   if (globerr == NOERR)
   {
      openQLalert(&ew);
      switch(globerr=err)
      {
         case NOHEAP : ZXprintf(ew,"\n NOT ENOUGH HEAP SPACE"); break;
         case MAXLIM : ZXprintf(ew,"\n MAXIMUM LIMIT REACHED"); break;
         default     : ZXprintf(ew,"\n UNKNOWN SYSTEM ERROR");
      }
      closeQLalert(ew);
      reset();
   }
}   /* end syserr */


/* reset input strings */
reset ()
{
   cstr[0] = '\0'; str = cstr; sp = 0;
   if_nst = 0; rt_ind = -1; clearlp(0);
}   /* end reset */


/* compares 2 strings */
streql (s,t)
   char *s, *t;
{
   while (*s == *t) { if (*s == '\0') return 1; s++; t++; }
   return 0;
}   /* end streql */


/* hash function */
hash(var)
    char var[];
{
   int i, h;

   h=0;
   for (i=0; var[i]!='\0'; i++) h+=var[i];

   return (h % MAXVAR);
}   /* end hash */


/*
 * QL specific commands begin here
 */


/* initializes the display */
initQLscr ()
{
   window(stdout,440,124,35,35);
   paper(stdout,7);
   ink(stdout,0);
   border(stdout,2,3);
   cursoff(stdout);
   flash(stdout, 0);
   mode(4);
}   /* end initQLscr */


/* opens alert box */
openQLalert (ew)
   int *ew;
{
   *ew = fopen("scr_150x28a300x40","w");
   paper(*ew,7); ink(*ew,0); border(*ew,2,3); cls(*ew);
}   /* end openQLalert */


/* closes alert box */
closeQLalert (ew)
   int ew;
{
   fgetc(stdin); border(ew,0,7); cls(ew); fclose(ew);
}   /* end closeQLalert */


ZXgets(fp, str, len)
   int fp; char str[]; int len;
{
   int i, j;
   char ch;
   
   row = 112; col = 0;
   ZXputc(fp, 'K', 7, 0);

   j = 0;
   while (1)
   {
      ch = getchar();

      if (ch == 194)
      {
         if (col > 0)
         {
            j--;
            ZXputc(fp, ' ', 0, 7);
            col = col - 8;
            ZXputc(fp, 'K', 7, 0);
         }
      }
      else if (ch == 10 || j == len-2)
      {
         str[j] = '\n';
         str[++j] = '\0';
         col = 0;
         for (i=0; i<j; i++)
         {
            ZXputc(fp, ' ', 0, 7);
            col = col + 8;
         }
         break;
      }
      else 
      {
         str[j++] = ch;
         ZXputc(fp, ch, 0, 7);
         col = col + 8;
         ZXputc(fp, 'K', 7, 0);
      }
   }

   row = col = 0;
   cls(fp);
}


ZXprintf(fp, str)
   int fp; char str[];
{
   int i;

   if (fp == stdout)
   {
      for (i=0; str[i]!='\0'; i++)
      {
         if (str[i] == 10)
         {
            downcursor(fp);
            continue;
         }
         else if (str[i] == '!')
            break;
         else if (str[i] == '\\')
         {
            ZXgraphics(fp, str[++i], str[++i]);
         }
         else if (str[i] == '%')
            ZXputc(fp, str[++i], 7, 0);
         else
            ZXputc(fp, str[i], 0, 7);

         adjustcursor(fp);
      }
   }
   else
      fprintf(fp, str);
}


ZXputc(fp, ch, ink, pap)
   int fp; char ch; int ink, pap;
{
   int i, x, y;

   if (fp == stdout)
   {
      if (ch == 10)
      {
         downcursor(fp);
      }
      else
      {
         ch = toupper(ch);
         block(fp, 8, 8, col, row, pap);

         i = 0;
         x = 0;
         for (y = 1; y < 7; y++)
         {
            while (x < 7)
            {
               if (ispixel(ch, i++))
                  block(fp, 1, 1, col+x, row+y, ink);
               x++;
            }
            x = 1;
         }
  
         for (x=2; x<5; x++)
            if (ispixel(ch, i++))
               block(fp, 1, 1, col+x, row+7, ink);
      }
   }
   else
      fputc(ch, fp);
}


ZXgraphics(fp, ch1, ch2)
   int fp; char ch1, ch2;
{
   switch (ch1)
   {
      case '#':
         if (ch2 == '#')
            { ZXgrey(fp, 1, 1, 0, 7); return; }
         break;
      case '~':
         if (ch2 == '~')
            { ZXgrey(fp, 1, 0, 0, 7); return; }
         break;
      case ',':
         if (ch2 == ',')
            { ZXgrey(fp, 0, 1, 0, 7); return; }
         break;
      case '@':
         if (ch2 == '@')
            { ZXgrey(fp, 1, 1, 7, 0); return; }
         break;
      case '!':
         if (ch2 == '!')
            { ZXgrey(fp, 1, 0, 7, 0); return; }
         break;
      case ';':
         if (ch2 == ';')
            { ZXgrey(fp, 0, 1, 7, 0); return; }
         break;
      case ' ':
         if (ch2 == ':')
            { ZXblock(fp, 0, 1, 0, 1); return; }
         if (ch2 == '\'')
            { ZXblock(fp, 0, 1, 0, 0); return; }
         if (ch2 == '.')
            { ZXblock(fp, 0, 0, 0, 1); return; }
         if (ch2 == ' ')
            { ZXputc(fp, ' ', 0, 7); return; }
               return;
         break;
      case ':':
         if (ch2 == ':')
            { ZXputc(fp, ' ', 7, 0); return; }
         if (ch2 == '\'')
            { ZXblock(fp, 1, 1, 1, 0); return; }
         if (ch2 == '.')
            { ZXblock(fp, 1, 0, 1, 1); return; }
         if (ch2 == ' ')
            { ZXblock(fp, 1, 0, 1, 0); return; }
         break;
      case '\'':
         if (ch2 == ':')
            { ZXblock(fp, 1, 1, 0, 1); return; }
         if (ch2 == '\'')
            { ZXblock(fp, 1, 1, 0, 0); return; }
         if (ch2 == '.')
            { ZXblock(fp, 1, 0, 0, 1); return; }
         if (ch2 == ' ')
            { ZXblock(fp, 1, 0, 0, 0); return; }
         break;
      case '.':
         if (ch2 == ':')
            { ZXblock(fp, 0, 1, 1, 1); return; }
         if (ch2 == '\'')
            { ZXblock(fp, 0, 1, 1, 0); return; }
         if (ch2 == '.')
            { ZXblock(fp, 0, 0, 1, 1); return; }
         if (ch2 == ' ')
            { ZXblock(fp, 0, 0, 1, 0); return; }
         break;
   }

   /* Question mark for unknown characters */
   ZXputc(fp, '?', 0, 7);
}


ZXgrey(fp, up, lo, ink, pap)
   int fp, up, lo, ink, pap;
{
   int x, y;

   block(fp, 8, 8, col, row, pap);

   if (up)
   {
      for (y=0; y<4; y++)
         for (x=0; x<8; x++)
            if ((x+y)%2 == 0)
               block(fp, 1, 1, col+x, row+y, ink);
   }
   if (lo)
   {
      for (y=4; y<8; y++)
         for (x=0; x<8; x++)
            if ((x+y)%2 == 0)
               block(fp, 1, 1, col+x, row+y, ink);
   }
}


ZXblock(fp, ul, ur, ll, lr)
   int fp, ul, ur, ll, lr;
{
   block(fp, 8, 8, col, row, 7);

   if (ul)
      block(fp, 4, 4, col, row, 0);
   if (ur)
      block(fp, 4, 4, col+4, row, 0);
   if (ll)
      block(fp, 4, 4, col, row+4, 0);
   if (lr)
      block(fp, 4, 4, col+4, row+4, 0);
}


downcursor(fp)
   int fp;
{
   col = 0;
   row = row + 8;
   if (row > LASTROW)
   {
      cls(fp);
      row = 0;
   }
}


adjustcursor(fp)
   int fp;
{
   col = col + 8;
   if (col > LASTCOL)
   {
      col = 0;
      row = row + 8;
      if (row > LASTROW)
      {
         cls(fp);
         row = 0;
      }
   }
}


ispixel(ch, i)
   char ch; int i;
{
   switch (ch)
   {
      case '"':
         return chQUO[i/8] & mask[i%8];
      case '#':
         return chPOU[i/8] & mask[i%8];
      case '$':
         return chDOL[i/8] & mask[i%8];
      case ':':
         return chCOL[i/8] & mask[i%8];
      case '(':
         return chOPA[i/8] & mask[i%8];
      case ')':
         return chCPA[i/8] & mask[i%8];
      case '>':
         return chGRE[i/8] & mask[i%8];
      case '<':
         return chLES[i/8] & mask[i%8];
      case '=':
         return chEQU[i/8] & mask[i%8];
      case '+':
         return chPLU[i/8] & mask[i%8];
      case '-':
         return chMIN[i/8] & mask[i%8];
      case '_':
         return chUND[i/8] & mask[i%8];
      case '*':
         return chAST[i/8] & mask[i%8];
      case '/':
         return chDIV[i/8] & mask[i%8];
      case ';':
         return chSEM[i/8] & mask[i%8];
      case ',':
         return chCOM[i/8] & mask[i%8];
      case '.':
         return chPER[i/8] & mask[i%8];
      case ' ':
         return chSPA[i/8] & mask[i%8];
      case '0':
         return ch0[i/8] & mask[i%8];
      case '1':
         return ch1[i/8] & mask[i%8];
      case '2':
         return ch2[i/8] & mask[i%8];
      case '3':
         return ch3[i/8] & mask[i%8];
      case '4':
         return ch4[i/8] & mask[i%8];
      case '5':
         return ch5[i/8] & mask[i%8];
      case '6':
         return ch6[i/8] & mask[i%8];
      case '7':
         return ch7[i/8] & mask[i%8];
      case '8':
         return ch8[i/8] & mask[i%8];
      case '9':
         return ch9[i/8] & mask[i%8];
      case 'A':
         return chA[i/8] & mask[i%8];
      case 'B':
         return chB[i/8] & mask[i%8];
      case 'C':
         return chC[i/8] & mask[i%8];
      case 'D':
         return chD[i/8] & mask[i%8];
      case 'E':
         return chE[i/8] & mask[i%8];
      case 'F':
         return chF[i/8] & mask[i%8];
      case 'G':
         return chG[i/8] & mask[i%8];
      case 'H':
         return chH[i/8] & mask[i%8];
      case 'I':
         return chI[i/8] & mask[i%8];
      case 'J':
         return chJ[i/8] & mask[i%8];
      case 'K':
         return chK[i/8] & mask[i%8];
      case 'L':
         return chL[i/8] & mask[i%8];
      case 'M':
         return chM[i/8] & mask[i%8];
      case 'N':
         return chN[i/8] & mask[i%8];
      case 'O':
         return chO[i/8] & mask[i%8];
      case 'P':
         return chP[i/8] & mask[i%8];
      case 'Q':
         return chQ[i/8] & mask[i%8];
      case 'R':
         return chR[i/8] & mask[i%8];
      case 'S':
         return chS[i/8] & mask[i%8];
      case 'T':
         return chT[i/8] & mask[i%8];
      case 'U':
         return chU[i/8] & mask[i%8];
      case 'V':
         return chV[i/8] & mask[i%8];
      case 'W':
         return chW[i/8] & mask[i%8];
      case 'X':
         return chX[i/8] & mask[i%8];
      case 'Y':
         return chY[i/8] & mask[i%8];
      case 'Z':
         return chZ[i/8] & mask[i%8];
      default:
         return chQUE[i/8] & mask[i%8];   }
}



/****************************** expr_c ******************************/


/* gets expression from command line */
getexpr (val_i,val_s)
   int *val_i; addr *val_s;
{
   return evalbool(val_i,val_s);
}   /* end getexpr */


/* parses parenthetical expression */
getsubexpr (val_i,val_s)
   int *val_i; addr *val_s;
{
   while (str[sp] == ' ') sp++;
   if (str[sp] == '(') sp++; else return NOVAL;
   if (getexpr(val_i,val_s) != NOERR) return BADARG;
   while (str[sp] == ' ') sp++;
   if (str[sp] == ')') sp++; else { free(*val_s); return BADCHAR; }

   return NOERR;
}   /* end getsubexpr */


/* gets a variable from command line */
getvar (var)
   char var[];
{
   int  i;
   char ch;

   while (str[sp] == ' ') sp++;
   if (!isalpha(str[sp])) return NOVAL;
   for (i=0; i<VARLEN; i++,sp++)
      if (isalpha(ch=str[sp]) || isdigit(ch) || (ch == '_'))
         var[i] = toupper(ch);
      else
         { if (ch == '$')
              { var[i++] = ch; sp++; if (i == VARLEN) break; }
           var[i] = '\0'; return NOERR; }
   return BADARG;
}   /* end getvar */


/* finds var in hashtable */
findvar(var,vp)
   char var[]; int *vp;
{
   int i, j;

   i = hash(var);
   for (j=i; j!=(i-1); j=(j+1)%MAXVAR)
      if (var_nam[j] == 0L)
         { *vp = j; return NOVAL; }
      else if (streql(var_nam[j],var))
         { *vp = j; return NOERR; }

   return MAXLIM;
}   /* end findvar */


/* sets a variable */
storevar (var,val_i,val_s)
   char var[]; int val_i; char val_s[];
{
   addr mem;
   int  i, len, err;

   if ((err=findvar(var,&i)) == MAXLIM)
      syserr(err);
   else
   {
      if (err == NOVAL)
         { memarg(&mem,strlen(var)+1);
           var_nam[i]=mem; strcpy(mem,var); }

      if (var[strlen(var)-1] != '$')
         var_num[i] = val_i;
      else
         { memarg(&mem,strlen(val_s)+1);
           if (err == NOERR) free(var_str[i]);
           var_str[i] = mem; strcpy(mem,val_s); }
   }
}   /* end storevar */


/* retrievs value of a variable */
fetchvar (var,val_i,val_s)
   char var[]; int *val_i; addr *val_s;
{
   int i, err;

   if ((err=findvar(var,&i)) == MAXLIM)
      syserr(err);
   else if (err == NOVAL)
      { memarg(val_s,1);
        *val_i = 0; strcpy(*val_s,""); }
   else if (var[strlen(var)-1] != '$')
      { memarg(val_s,NUMLEN);
        *val_i = var_num[i]; itoa(*val_i,*val_s); }
   else
      { memarg(val_s,strlen(var_str[i])+1);
        strcpy(*val_s,var_str[i]); *val_i = atoi(*val_s); }
}   /* end fetchvar */


/* gets number from command line */
getnum (val_i,val_s)
   int *val_i; addr *val_s;
{
   int i;

   while (str[sp] == ' ') sp++;
/*
   if (!isdigit(str[sp]) && ((str[sp]!='-') || !isdigit(str[sp+1]))) return NOVAL;
*/
   if (!isdigit(str[sp])) return NOVAL;
   i=0; memarg(val_s,NUMLEN);
   if (str[sp] == '-') setchp(*val_s,i++,str[sp++]);
   while (isdigit(str[sp]))
      setchp(*val_s,i++,str[sp++]);
   setchp(*val_s,i,'\0');
   *val_i = atoi(*val_s);

   return NOERR;
}   /* end getnum */


/* gets quoted string from command line */
getstr (val_i,val_s)
   int *val_i; addr *val_s;
{
   int i, tp;

   while (str[sp] == ' ') sp++;
   if (str[sp] == '"') tp = ++sp; else return NOVAL;
   while(str[tp]!='"') if (str[tp]!='\n') tp++; else return BADCHAR;
   memarg(val_s,tp-sp+1);
   for (i=0; str[sp]!='"'; i++) setchp(*val_s,i,str[sp++]);
   setchp(*val_s,i,'\0'); sp++;
   *val_i = atoi(*val_s);

   return NOERR;   
}   /* end getstr */


/* parses substring expression */
getsubstr (tmp_i,tmp_s,val_i,val_s)
   int *tmp_i; addr *tmp_s;
   int *val_i; addr *val_s;
{
   int err, from, to, i, j;

   if ((err=getrange(&from,&to,strlen(*tmp_s))) == NOVAL)
      { *val_i=*tmp_i; *val_s=*tmp_s; return NOERR; }
   else if (err != NOERR)
      { free(*tmp_s); return BADARG; }
      
   memarg(val_s,(to-from+2));
   j=0; for (i=(from-1); i<to; i++) setchp(*val_s,j++,getchp(*tmp_s,i));
   setchp(*val_s,j,'\0');  *val_i = atoi(*val_s);

   free(*tmp_s); return NOERR;
}   /* end getsubstr */


/* gets range of string */
getrange (from,to,len)
   int *from, *to, len;
{
   char *tmp;
   int  noTO;

   if (getsym("(") != NOERR) return NOVAL;

   if (noTO=(getkey("TO") != NOERR))
      { if (getexpr(from,&tmp) ==NOERR) free(tmp); else return BADARG; }
   else
      *from = 1;

   if (!noTO || (getkey("TO") == NOERR))
      { if (getexpr(to,&tmp) == NOERR) free(tmp); else *to = len; }
   else
      *to = *from;

   if (getsym(")") != NOERR) return BADARG;

   if (*from < 1) *from = 1;
   if (*to > len) *to = len;
   if (*to < *from) *to = *from-1;
   return NOERR;
}   /* end getrange */


/* gets value from command line */
getval (val_i,val_s)
   int *val_i; addr *val_s;
{
   addr tmp_s;
   int  tmp_i, err;
   char var[VARLEN];

   if ((err=getsubexpr(val_i,val_s)) != NOVAL)
      return err;
   else if ((err=evalfun(val_i,val_s)) != NOVAL)
      return err;
   else if ((err=getnum(val_i,val_s)) != NOVAL)
      return err;
   else if ((err=getstr(&tmp_i,&tmp_s)) != NOVAL)
      { if (err != NOERR) return err;
        return getsubstr(&tmp_i,&tmp_s,val_i,val_s); }
   else if (getvar(var) == NOERR)
      { fetchvar(var,&tmp_i,&tmp_s);
        return getsubstr(&tmp_i,&tmp_s,val_i,val_s); }

   return BADARG;
}   /* end getval */


/* gets {chr$(),len()} functions and unary {-,not} expressions */
evalfun (val_i,val_s)
   int *val_i; addr *val_s;
{
   while (str[sp] == ' ') sp++;
   if (getsym("-") == NOERR) {
      if (getval(val_i,val_s) != NOERR) return BADARG;
      *val_i = -*val_i;
   }
   else if (getsym("CHR$") == NOERR) {
      if (getsubexpr(val_i,val_s) != NOERR) return BADARG;
      *val_i %= 256; setchp(*val_s,0,*val_i); if (*val_i) setchp(*val_s,1,'\0');
      return NOERR;
   }
   else if (getsym("CODE") == NOERR) {
      if (getsubexpr(val_i,val_s) != NOERR) return BADARG;
      *val_i = getchp(*val_s,0);
   }
   else if (getsym("LEN") == NOERR) {
      if (getsubexpr(val_i,val_s) != NOERR) return BADARG;
      *val_i = strlen(*val_s);
   }
   else if (getkey("NOT") == NOERR) {
      if (evaleql(val_i,val_s) != NOERR) return BADARG;
      *val_i = !*val_i;
   }
   else return NOVAL;

   free(*val_s); memarg(val_s,NUMLEN); itoa(*val_i,*val_s);
   return NOERR;
}   /* end evalfun */


/* parses & expressions */
evalcat (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i, tmplen;
   addr right_s, tmp_s;

   if (getval(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case '&' : {
            sp++; tmplen = strlen(*left_s);
            memarg(&tmp_s,tmplen+1);
            strcpy(tmp_s,*left_s); free(*left_s);
            if (getval(&right_i,&right_s) != NOERR) { free(tmp_s); return BADARG; }
            memarg(left_s,tmplen+strlen(right_s)+1);
            strcpy(*left_s,tmp_s); free(tmp_s);
            strcat(*left_s,right_s); free(right_s);
            *left_i = atoi(*left_s);
            break;
         }
         default  : return NOERR;
      }
   }
}   /* end evalcat */


/* parses ^ expressions */
evalpow (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i, pow, i;
   addr right_s;

   if (evalcat(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case '^' : {
            sp++;
            if (evalcat(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            pow = 1;
            for (i=0; i<right_i; i++)
               pow *= *left_i;

            *left_i = pow;
            break;
         }
         default  : {
            return NOERR;
         }
      }
      free (*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}   /* end evalpow */


/* parses {*,/} expressions */
evalmul (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i;
   addr right_s;

   if (evalpow(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case '*' : {
            sp++;
            if (evalpow(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            *left_i = *left_i * right_i;
            break;
         }
         case '/' : {
            sp++;
            if (evalpow(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            *left_i = *left_i / right_i;
            break;
         }
         default  : {
            return NOERR;
         }
      }
      free(*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}   /* end evalmul */


/* parses {+,-} expressions */
evaladd (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i;
   addr right_s;

   if (evalmul(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case '+' : {
            sp++;
            if (evalmul(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            *left_i = *left_i + right_i;
            break;
         }
         case '-' : {
            sp++;
            if (evalmul(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            *left_i = *left_i - right_i;
            break;
         }
         default  : {
            return NOERR;
         }
      }
      free(*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}   /* end evaladd */


/* parses {=,>,<,<>,<=,>=} expressions */
evaleql (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i;
   addr right_s;

   if (evaladd(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      if (getsym("=") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if ((*left_i) || (right_i))
            *left_i = *left_i == right_i;
         else
            *left_i = (strcmp(*left_s,right_s) == 0);
      }
      else if (getsym("<=") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if ((*left_i) || (right_i))
            *left_i = *left_i <= right_i;
         else
            *left_i = (strcmp(*left_s,right_s) <= 0);
      }
      else if (getsym("<>") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if ((*left_i) || (right_i))
            *left_i = *left_i != right_i;
         else
            *left_i = (strcmp(*left_s,right_s) != 0);
      }
      else if (getsym("<") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if ((*left_i) || (right_i))
            *left_i = *left_i < right_i;
         else
            *left_i = (strcmp(*left_s,right_s) < 0);
      }
      else if (getsym(">=") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if ((*left_i) || (right_i))
            *left_i = *left_i >= right_i;
         else
            *left_i = (strcmp(*left_s,right_s) >= 0);
      }
      else if (getsym(">") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if ((*left_i) || (right_i))
            *left_i = *left_i > right_i;
         else
            *left_i = (strcmp(*left_s,right_s) > 0);
      }
      else {
         return NOERR;
      }
      free(*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}   /* end evaleql */


/* parses {and,or} expressions */
evalbool (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i;
   addr right_s;

   if (evaleql(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      if (getkey("AND") == NOERR) {
         if (evaleql(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         *left_i = *left_i && right_i;
      }
      else if (getkey("OR") == NOERR) {
         if (evaleql(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         *left_i = *left_i || right_i;
      }
      else {
         return NOERR;
      }
      free(*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}   /* end evalbool */

